/*
* Priority inversion test
* Author: Sergey Reva (rs_rlab@mail.ru)
* http://rs-rlab.narod.ru
* 
* Ported from QNX version,
* which created by Olej (olej@front.ru)
* 
*/

#include <u.h>
#include <libc.h>
#include "slock.h"

enum {
	false,
	true,
	THRNUM = 3,
	NCKL = 10,
	LINOUT = 5,
	LMSG = 10,
	BUFLEN = THRNUM * NCKL * LMSG + 1,
};

void workfun(ulong n)
{
	ulong i;
	double f;

	for(i = 0; i < n; i++ )
		f = sqrt( 1. + sqrt( (double) rand() ) );

	USED(f);
}

char buf[ BUFLEN ];
uint ind = 0;
long nrep = 500000;
int ndebug;

SLock mutex;
Lock aav;

uint atomic_add_value(uint val)
{
	uint v;
	lock(&aav);
	ind+=val;
	v=ind;
	unlock(&aav);
	return v;
}

void thrfunc(void *p)
{
	int i,t = (int)p;	
	char nt = t + '1', s[ LMSG +1 ];

	if(t!=1)
		slock(&mutex);

	sleep(1);

	for(i = 0; i < NCKL; i++ )
	{
		if( ndebug > 0 )
			sprint(s,"%c[%d] ",nt, getpri(getpid()));
		else 
			sprint(s,"%c",nt);

		strncpy(buf + ind,s,strlen(s));

		atomic_add_value(strlen(s));
		workfun(nrep);
	}

	if(t!=1)
		sunlock(&mutex);
};

void pthread_create(void* p,int pri)
{
	switch(rfork(RFPROC|RFMEM))
	{
	case -1:
		abort();
	case 0:
		setpri(getpid(),pri);
		thrfunc(p);
		exits(0);
	default:
		return;
	}
}

int main( int argc, char *argv[] )
{
	Waitmsg *wm;
	int i;

	rfork(RFREND);

	print("inverse test, Plan 9 with SLock library, vers.1.04\n");
	
	ARGBEGIN
	{
	case 'v': 
		ndebug++; 
		break;
	case 't':
		nrep = atol(ARGF());
		break;
	case 'h':
	default:
		print("usage [-v] -t times");
		exits(0);
	} ARGEND

	print("repeating number =  %d, debug level = %d\n",nrep,ndebug);

	for(i = 0; i<THRNUM; i++)
		pthread_create((void*)i,11+i);

	for(;;)
	{
		wm=wait();
		if (wm!=nil)
			free(wm);
		else
			break;
	}

	buf[ind] = '\0';
	print("%s\n",buf);

	exits(0);
}
