/*
* Test program for slock library
* Author: Sergey Reva (rs_rlab@mail.ru)
* http://rs-rlab.narod.ru
*/

#include <u.h>
#include <libc.h>
#include "slock.h"

static struct
{
	SLock;
	char d[1024];
	int pos;
} buf;

enum {
	MaxThread=6,	//maximum is 10 
	
	SleepPerThrd=2,
	SleepZeroThrd=1000,
};
//Process:	0	1	2	3	4	5	6	7	8	9
int prio[]={	10,	12,	12,	15,	13,	12,	16,	17,	18,	19};
/*
*
* This table show priority for processes
* Output should be in the order of priority
* except 0 - first obtain object, and hold it 1 sec
* except 1 - it obtain object only when it be free
*
* By default (for 6 process):
* 0 - 10 always first
* 1 - 12 probably last (sixth) but can catch the moment in middle
* 2 - 12 as less prioritized, but third come got fourth place
* 3 - 15 as most prioritized should appeared second
* 4 - 13 as second most pri. should got third place
* 5 - 12 as less prioritized, and last come got fifth place
*
* 034251
*
*/

void proc(char i)
{
	char n;

	switch(rfork(RFPROC|RFMEM))
	{
	case 0:
		setpri(getpid(),prio[i]);
		break;
	case 1:
		print("can't fork\n");
		abort();
	default:
		return;
	}

	n=i+'0';

	switch(n)
	{
	case '0':		//This process capture object for long time
				//this time should be enough to form queue
		slock(&buf);
		buf.d[buf.pos]=n;
		buf.pos++;
		sleep(SleepZeroThrd);
		sunlock(&buf);
		break;
	case '1':		//This process try capture object in nonblock
	case1:
		if (canslock(&buf))
		{
			buf.d[buf.pos]=n;
			buf.pos++;	
			sunlock(&buf);
			break;
		}
		else
		{
			sleep(0);
			goto case1;
		}
		break;
	default:		//Simple worker
		slock(&buf);
		buf.d[buf.pos]=n;
		buf.pos++;
		sunlock(&buf);
		break;
	}

	exits(0);
}

void main(int,char**)
{
	char i;
	Waitmsg *wm;

	for (i=0;i<MaxThread;i++)
	{
		proc(i);
		sleep(SleepPerThrd);
	}

	for(;;)
	{
		wm=wait();
		if (wm!=nil)
			free(wm);
		else
			break;
	}	

	buf.d[buf.pos]='\n';
	buf.d[buf.pos+1]=0;
	print(buf.d);

	exits(0);
}