/*
* SLock library
* Provide priority inheritance for avoiding priority inversion
* Author: Sergey Reva (rs_rlab@mail.ru)
* http://rs-rlab.narod.ru
*
*/

#include <u.h>
#include <libc.h>
#include "slock.h"

#define DBGPRINT if(0)print

#define poff 28+28+12+6*12+12	//base priority offset

static struct {
	Lock;
	QI i[1024];
} qpl;

void setpri(int pid,int pri)
{
	int fd;
	char file[256];

	snprint(file,256,"#p/%d/ctl",pid);
	
	if ((fd=open(file,OWRITE))<0)
		return;
	fprint(fd,"pri %d\n",pri);
	close(fd);
}

int getpri(int pid)
{
	int fd;
	char file[256];
	char pri[12];

	snprint(file,256,"#p/%d/status",pid);
	
	if ((fd=open(file,OREAD))<0)
		return 0;
	pread(fd,pri,11,poff);
	close(fd);

	pri[11]=0;
	return atoi(pri);
}

QI* getqi(void)
{
	int i;

	lock(&qpl);

	for(i=0;i<1024;i++)
		if (!(qpl.i[i].flags & FUsed))
		{
			qpl.i[i].flags=FUsed;
			unlock(&qpl);
			return &qpl.i[i];
		}

	unlock(&qpl);
	return nil;

	//return malloc(sizeof(QI)); - this is also way
}

void freeqi(QI *i)
{
	i->flags=0;
	//free(i); - see getqi for comment
}

void slock(SLock* l)
{
	QI *qi,*qp;

	if ((qi=getqi())==nil)
		abort();

	qi->pid=getpid();
	qi->pri=getpri(qi->pid);
	qi->next=nil;

	lock(&l->Lock);

	if (l->first!=nil)
	{
		for(qp=l->first; qp->next!=nil; qp=qp->next)
			if (qp->next->pri < qi->pri)
				break;

		qi->next=qp->next;
		qp->next=qi;

		DBGPRINT("SL: Proc %d join to queue after proc %d\n",qi->pid,qp->pid);

		if 	((qi==l->first->next) 		//if we at top of queue
			&& (qi->pri > l->first->pri))	//and have greater priority
		{							//then boost current owner
			l->first->flags|=FBoost;
			setpri(l->first->pid,qi->pri);
			DBGPRINT("SL: Proc %d boost proc %d to %d\n",qi->pid,l->first->pid,qi->pri);
		}

		unlock(&l->Lock);

		while(rendezvous((ulong)qi,0x01234567)!=0x76543210)
			;
		DBGPRINT("SL: Proc %d got lock\n",qi->pid);
		return;
	}

	DBGPRINT("SL: Proc %d join to queue at begin\n",qi->pid);
	l->first=qi;
	unlock(&l->Lock);
}

void sunlock(SLock* l)
{
	QI* qi;

	lock(&l->Lock);

	qi=l->first;
	
	if (qi->flags & FBoost)
	{
		setpri(qi->pid,qi->pri);
		DBGPRINT("SL: Proc %d slow-down\n",qi->pid);
	}
	
	if (qi->next!=nil)
	{
		DBGPRINT("SL: Proc %d give lock to proc %d\n",qi->pid,qi->next->pid);
		while(rendezvous((ulong)qi->next,0x76543210)!=0x01234567)
			;
	}

	DBGPRINT("SL: Proc %d leave lock\n",qi->pid);
	l->first=l->first->next;
	freeqi(qi);
	unlock(&l->Lock);
}

int canslock(SLock* l)
{
	QI *qi;

	if (!canlock(&l->Lock))
		return 0;

	if (l->first!=nil)
	{
		unlock(&l->Lock);
		return 0;
	}

	if ((qi=getqi())==nil)
		abort();

	qi->pid=getpid();
	qi->pri=getpri(qi->pid);
	qi->next=nil;

	l->first=qi;
	unlock(&l->Lock);
	return 1;
}
